module lv_conv2d_f90
! very similar to lv_conv2d_mod, except that we take filter coef as argument in conv
! so that we can use GEE solvers


 use sep
 use lv_conv2d_mod

 implicit none

 real, dimension(:,:), pointer:: modl
 type(lv_filter2d), pointer	:: flter !=> must be initialized
 integer					:: nm1, nm2

contains
  
  subroutine conv2d_init_m(modl_in)
    real, dimension(:,:), allocatable, target, intent(in) :: modl_in
   ! call lv_conv2d_init_m(modl_in)
   modl =>modl_in
  end subroutine conv2d_init_m
  
  subroutine conv2d_init_f(flt_in)
    type(lv_filter2d), target, intent(in) :: flt_in
    flter => flt_in
   ! call lv_conv2d_init_f(flt_in)
  end subroutine conv2d_init_f
  
  subroutine conv2d_init(nd1_in,nd2_in)
    integer, intent(in) :: nd1_in,nd2_in
    nm1 = nd1_in
    nm2 = nd2_in
  end subroutine conv2d_init
  
  integer function conv2d_m_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call conv2d_m(adj,add,coef,data)
     stat=0
  end function   conv2d_m_op
  
  subroutine conv2d_m(adj,add,coef,data)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:),allocatable		 :: dat 
     flter%coef2=reshape(coef,shape(flter%coef2))
     dat=reshape(data,(/nm1, nm2/))
     call lv_conv2d_m(adj,add,flter,dat)
  end subroutine conv2d_m
  
  
  integer function DK_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call DK(adj,add,coef,data)
     stat=0
  end function   DK_op
  
  
   subroutine DK(adj,add,coef,data)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:),allocatable		 :: dat
     real, dimension(:,:,:),allocatable			 :: coeftmp
     
    allocate(dat(nm1,nm2),coeftmp(size(flter%coef2,1),size(flter%coef2,2),size(flter%coef2,3))) 
    call lv_conv2d_init_m(modl) 
        dat=reshape(data,shape(dat))
          flter%coef2=reshape(coef,shape(flter%coef2))
     if (adj) then
  
      call lv_conv2d_m(adj,add,flter,dat)
      flter%coef2(1,:,:)=0.
      coef=reshape(flter%coef2,shape(coef))
     else
      flter%coef2(1,:,:)=0.
      call lv_conv2d_m(adj,add,flter,dat)
      data=reshape(dat,shape(data))   
	 end if
   end subroutine  DK
   
   integer function DK2_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call DK2(adj,add,coef,data)
     stat=0
   end function  DK2_op
  
  
   subroutine DK2(adj,add,coef,data)  ! D(I-K)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:),allocatable		 :: dat 
     real, dimension(:,:),allocatable			 :: coeftmp
     
     call lv_conv2d_init_m(modl)
     
     allocate(coeftmp(size(flter%coef2,2),size(flter%coef2,3)),dat(nm1,nm2))
      dat=reshape(data,shape(dat))
      flter%coef2=reshape(coef,shape(flter%coef2))
     if (adj) then
      call lv_conv2d_m(adj,add,flter,dat)
      coeftmp=flter%coef2(1,:,:)
      flter%coef2=0.
      flter%coef2(1,:,:)=coeftmp
      coef=reshape(flter%coef2,shape(coef))
     else
      coeftmp=flter%coef2(1,:,:)        
      flter%coef2=0. 
      flter%coef2(1,:,:)=coeftmp
      call lv_conv2d_m(adj,add,flter,dat)
      data=reshape(dat,shape(data))
     end if
   end subroutine  DK2
  
  integer function conv2d_f_op(adj,add,mod,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: mod, data
     call conv2d_f(adj,add,mod,data)
     stat=0
  end function   conv2d_f_op
  
  subroutine conv2d_f(adj,add,mod,data)
     logical, intent(in)  	     :: adj, add
     real, dimension(nm1*nm2)		 	 :: mod,data
     real, dimension(:,:),allocatable		 :: modd,dat  
     allocate(modd(nm1,nm2),dat(nm1,nm2))
     
     call lv_conv2d_init_f(flter)
     
     modd=reshape(mod,shape(modd))
     dat=reshape(data,shape(dat))
     call lv_conv2d_f(adj,add,modd,dat)    
     mod=reshape(modd,shape(mod))
     data=reshape(dat,shape(data))
     
     call lv_conv2d_close_f()
  end subroutine conv2d_f
  
  subroutine conv2d_close_m()
    call lv_conv2d_close_m()   
  end subroutine conv2d_close_m
  
  subroutine conv2d_close_f()
    call lv_conv2d_close_f()     
  end subroutine conv2d_close_f
  
  integer function id_op(adj,add,mod,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: mod, data
     
     if (adj) then
      if (.not. add) mod=0.
      mod=mod+data
     else
      if (.not. add) data=0.
      data=data+mod
     end if
     
     stat=0
  end function   id_op

end module
