module lv_conv2d_mod
  
  use lv_filters_mod
  
  
  implicit none
  real, dimension(:,:), pointer:: model
  class(lv_filter), pointer:: flt !might be 1d or 2d

contains

  subroutine lv_conv2d_init_m(modl_in)
    real, dimension(:,:), target, intent(in) :: modl_in
    model=>modl_in
  end subroutine
  
  subroutine lv_conv2d_init_f(flt_in)
    class(lv_filter), target, intent(in) :: flt_in
    flt => flt_in
  end subroutine
  
  subroutine lv_conv2d_m(adj,add,flt2d, data)
     logical, intent(in)     :: adj, add
     class (lv_filter)		 :: flt2d
     real, dimension(:,:)		 :: data   
     if (adj) then
       if (.not. add) then
        select type (flt2d)
        type is (lv_filter1d)
        flt2d%coef1=0.
        type is (lv_filter2d)
        flt2d%coef2=0.
        end select
       end if
       call adjoint2d_m(model,flt2d,data)
     else
       if (.not. add) data=0.
       call forward2d(model,flt2d,data)
     end if  
  end subroutine lv_conv2d_m
  
  subroutine lv_conv2d_f(adj,add,mod, data)
     logical, intent(in)     :: adj, add
     real, dimension(:,:)		 :: mod, data
     
     if (adj) then
       if (.not. add) mod=0.
       call adjoint2d_f(mod,flt,data)
     else
       if (.not. add) data=0.
       call forward2d(mod,flt,data)
     end if  
  end subroutine lv_conv2d_f  
  
  subroutine lv_conv2d_close_m()
    nullify(model)
  end subroutine
  
  subroutine lv_conv2d_close_f()
    nullify(flt)
  end subroutine
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine forward2d(m,f,d) !output d=f*m
     real, dimension(:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:)		:: d
     
     integer        :: i,j,id1,id2,k1,k2, nlag
     real, allocatable :: b0(:),d1(:),bin(:), b1(:), d02(:), d12(:)
     
     select type (f)
     
     type is (lv_filter1d)
     
!     
     
     type is (lv_filter2d)

     nlag=size(f%lag,2)
     allocate(b0(nlag),bin(nlag),d1(nlag),b1(nlag),d02(nlag),d12(nlag)) 
      
     do i=1,size(f%b,2)  ! i-th patch
       k1=mod(i-1,(size(f%coef2,2)-1))+1
       k2=(i-1)/(size(f%coef2,2)-1)+1
       b0=f%coef2(:,k1,k2)
       b1=f%coef2(:,k1+1,k2)
       d02=(f%coef2(:,k1,k2+1)-b0)/(f%e(2,i)-f%b(2,i)+1)
       d12=(f%coef2(:,k1+1,k2+1)-b1)/(f%e(2,i)-f%b(2,i)+1)
       do id2=f%b(2,i),f%e(2,i)  ! slow
         bin=b0
         d1=(b1-b0)/(f%e(1,i)-f%b(1,i)+1)
         do id1=f%b(1,i),f%e(1,i) ! Fast axis
         ! TODO: mask check (mask is 2D)
           if( associated(f%mis)) then
             if(f%mis(id1,id2)) then
               cycle
             endif
           endif
           do j = 1,nlag
             d(id1+f%lag(1,j),id2+f%lag(2,j))=d(id1+f%lag(1,j),id2+f%lag(2,j))+bin(j)*m(id1,id2)
           end do       
           bin=bin+d1       
         end do   
         b0=b0+d02
         b1=b1+d12  
       end do
     end do 
     
     deallocate(b0,b1,bin,d1,d02,d12)
     
     end select
     
  end subroutine

  subroutine adjoint2d_f(m,f,d)
     real, dimension(:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:)		:: d     
     integer     			    :: i,j,id1,id2,k1,k2, nlag
     real, allocatable :: b0(:),d1(:),bin(:), b1(:), d02(:), d12(:)
     
     select type (f)
     
     type is (lv_filter1d)
     
     !
     !
     !

     
     type is (lv_filter2d)

     nlag=size(f%lag,2)
     allocate(b0(nlag),bin(nlag),d1(nlag),b1(nlag),d02(nlag),d12(nlag)) 
      
     do i=1,size(f%b,2)  ! i-th patch
       k1=mod(i-1,(size(f%coef2,2)-1))+1
       k2=(i-1)/(size(f%coef2,2)-1)+1
       b0=f%coef2(:,k1,k2)
       b1=f%coef2(:,k1+1,k2)
       d02=(f%coef2(:,k1,k2+1)-b0)/(f%e(2,i)-f%b(2,i)+1)
       d12=(f%coef2(:,k1+1,k2+1)-b1)/(f%e(2,i)-f%b(2,i)+1)
       do id2=f%b(2,i),f%e(2,i)  ! slow
         bin=b0
         d1=(b1-b0)/(f%e(1,i)-f%b(1,i)+1)
         do id1=f%b(1,i),f%e(1,i) ! Fast axis
         ! TODO: mask check (mask is 2D)
         do j = 1,nlag
             m(id1,id2)=m(id1,id2)+bin(j)*d(id1+f%lag(1,j),id2+f%lag(2,j))
           end do       
           bin=bin+d1       
         end do   
         b0=b0+d02
         b1=b1+d12  
       end do
     end do 
     
     deallocate(b0,b1,bin,d1,d02,d12)
     
     end select

  end subroutine adjoint2d_f
  
  subroutine adjoint2d_m(m,f,d)
     real, dimension(:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:)		:: d
     
     integer        :: i,j,id1,id2,k1,k2, nlag
     double precision         :: invn1, invn2
     real, allocatable :: b0(:),d1(:),bin(:), b1(:), d02(:), d12(:)
     real :: wt,x
     
     select type (f)
     
     type is (lv_filter1d)
     
     !
     !
     
     type is (lv_filter2d)

     nlag=size(f%lag,2)
     allocate(b0(nlag),bin(nlag),d1(nlag),b1(nlag),d02(nlag),d12(nlag)) 
     
     do i=1,size(f%b,2)  ! i-th patch
       invn1=1./(f%e(1,i)-f%b(1,i)+1)
       invn2=1./(f%e(2,i)-f%b(2,i)+1)
       k1=mod(i-1,(size(f%coef2,2)-1))+1
       k2=(i-1)/(size(f%coef2,2)-1)+1
       d02=0
       d12=0
       b0=0
       b1=0
    !   do id2=f%b(2,i),f%e(2,i)  ! slow
        do id2=f%e(2,i),f%b(2,i),-1
         bin=0
         d1=0
      !   wt=0
         !do id1=f%b(1,i),f%e(1,i) ! Fast axis
          do id1=f%e(1,i),f%b(1,i),-1
         ! TODO: mask check (mask is 2D)
 if( associated(f%mis)) then
                if(f%mis(id1,id2)) then
                             cycle
                endif
         endif 
          
         do j = 1,nlag
                x= m(id1,id2)*d(id1+f%lag(1,j),id2+f%lag(2,j))
                d1(j)=d1(j)+bin(j)
            	bin(j)=x+bin(j)
            !	d1(j)=d1(j)+wt*x
           end do
          !      wt=wt+1
         end do
           d02=d02+b0
           d12=d12+b1
           b0 = b0+bin-d1*invn1
           b1 = b1+d1*invn1              
       end do
      !  d02=d02+b0
      !  d12=d12+b1
       
       f%coef2(:,k1,k2) = f%coef2(:,k1,k2) +b0 -invn2* d02
       f%coef2(:,k1+1,k2) =  f%coef2(:,k1+1,k2) +b1 -invn2 * d12
       f%coef2(:,k1,k2+1) = f%coef2(:,k1,k2+1) + invn2 * d02
       f%coef2(:,k1+1,k2+1) = f%coef2(:,k1+1,k2+1) + invn2 * d12
       
     end do 
     
     deallocate(b0,b1,bin,d1,d02,d12)
     
     end select
   
  end subroutine adjoint2d_m



end module
