module lv_conv3d_f90
! very similar to lv_conv3d_mod, except that we take filter coef as argument in conv
! so that we can use GEE solvers


 use sep
 use lv_conv3d_mod

 implicit none

 real, dimension(:,:,:), pointer:: modl
 type(lv_filter3d), pointer	:: flter !=> must be initialized
 integer					:: nm1, nm2, nm3

contains
  
  subroutine conv3d_init_m(modl_in)
    real, dimension(:,:,:), allocatable, target, intent(in) :: modl_in
   ! call lv_conv2d_init_m(modl_in)
   modl =>modl_in
  end subroutine conv3d_init_m
  
  subroutine conv3d_init_f(flt_in)
    type(lv_filter3d), target, intent(in) :: flt_in
    flter => flt_in
   ! call lv_conv2d_init_f(flt_in)
  end subroutine conv3d_init_f
  
  subroutine conv3d_init(nd1_in,nd2_in,nd3_in)
    integer, intent(in) :: nd1_in,nd2_in,nd3_in
    nm1 = nd1_in
    nm2 = nd2_in
    nm3 = nd3_in
  end subroutine conv3d_init
  
  integer function conv3d_m_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call conv3d_m(adj,add,coef,data)
     stat=0
  end function   conv3d_m_op
  
  subroutine conv3d_m(adj,add,coef,data)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:,:),allocatable		 :: dat
     
     allocate(dat(nm1,nm2,nm3))
      
     flter%coef3=reshape(coef,shape(flter%coef3))
     dat=reshape(data,shape(dat))
     call lv_conv3d_m(adj,add,flter,dat)
     coef=reshape(flter%coef3, shape(coef))
     data=reshape(dat,shape(data))
  end subroutine conv3d_m
  
  
  integer function DK3d_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call DK3d(adj,add,coef,data)
     stat=0
  end function   DK3d_op
  
  
   subroutine DK3d(adj,add,coef,data)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:,:),allocatable		 :: dat
 
    allocate(dat(nm1,nm2,nm3))
    call lv_conv3d_init_m(modl) 
    dat=reshape(data,shape(dat))
    flter%coef3=reshape(coef,shape(flter%coef3))
    if (adj) then
      call lv_conv3d_m(adj,add,flter,dat)
      flter%coef3(1,:,:,:)=0.
      coef=reshape(flter%coef3,shape(coef))
     else
      flter%coef3(1,:,:,:)=0.
      call lv_conv3d_m(adj,add,flter,dat)
      data=reshape(dat,shape(data))   
	 end if
	 
   end subroutine  DK3d
   
   integer function DK23d_op(adj,add,coef,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: coef, data
     call DK23d(adj,add,coef,data)
     stat=0
   end function  DK23d_op
  
  
   subroutine DK23d(adj,add,coef,data)  ! D(I-K)
     logical, intent(in)  	  				 :: adj, add
     real, dimension(:)						 :: coef
     real, dimension(:)						 :: data  
     real, dimension(:,:,:),allocatable		 :: dat 
     real, dimension(:,:,:),allocatable		 :: coeftmp
     
     call lv_conv3d_init_m(modl)
     
     allocate(coeftmp(size(flter%coef3,2),size(flter%coef3,3),size(flter%coef3,4)),dat(nm1,nm2,nm3))
     dat=reshape(data,shape(dat))
     flter%coef3=reshape(coef,shape(flter%coef3))
     if (adj) then
       call lv_conv3d_m(adj,add,flter,dat)
       coeftmp=flter%coef3(1,:,:,:)
       flter%coef3=0.
       flter%coef3(1,:,:,:)=coeftmp
       coef=reshape(flter%coef3,shape(coef))
     else
      coeftmp=flter%coef3(1,:,:,:)        
      flter%coef3=0. 
      flter%coef3(1,:,:,:)=coeftmp
      call lv_conv3d_m(adj,add,flter,dat)
      data=reshape(dat,shape(data))
     end if
   end subroutine  DK23d
  
  integer function conv3d_f_op(adj,add,mod,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: mod, data
     call conv3d_f(adj,add,mod,data)
     stat=0
  end function   conv3d_f_op
  
  subroutine conv3d_f(adj,add,mod,data)
     logical, intent(in)  	     :: adj, add
     real, dimension(nm1*nm2*nm3)		 	 :: mod,data
     real, dimension(:,:,:),allocatable		 :: modd,dat  
     allocate(modd(nm1,nm2,nm3),dat(nm1,nm2,nm3))
     
     call lv_conv3d_init_f(flter)
     
     modd=reshape(mod,shape(modd))
     dat=reshape(data,shape(dat))
     call lv_conv3d_f(adj,add,modd,dat)    
     mod=reshape(modd,shape(mod))
     data=reshape(dat,shape(data))
     
     call lv_conv3d_close_f()
  end subroutine conv3d_f
  
  subroutine conv3d_close_m()
    call lv_conv3d_close_m()   
  end subroutine conv3d_close_m
  
  subroutine conv3d_close_f()
    call lv_conv3d_close_f()     
  end subroutine conv3d_close_f
  
  integer function id3d_op(adj,add,mod,data) result(stat)
     logical, intent(in)  	     :: adj, add
     real, dimension(:)			 :: mod, data
     
     if (adj) then
      if (.not. add) mod=0.
      mod=mod+data
     else
      if (.not. add) data=0.
      data=data+mod
     end if
     
     stat=0
  end function   id3d_op

end module
