module lv_conv3d_mod
  
  use lv_filters_mod
  
  implicit none
  real, dimension(:,:,:), pointer:: model
  class(lv_filter), pointer:: flt

contains

  subroutine lv_conv3d_init_m(modl_in)
    real, dimension(:,:,:), target, intent(in) :: modl_in
    model=>modl_in
  end subroutine lv_conv3d_init_m
  
  subroutine lv_conv3d_init_f(flt_in)
    class(lv_filter), target, intent(in) :: flt_in
    flt => flt_in
  end subroutine lv_conv3d_init_f
  
  subroutine lv_conv3d_m(adj,add,flt3d, data)
     logical, intent(in)     :: adj, add
     class (lv_filter)		 :: flt3d
     real, dimension(:,:,:)		 :: data   
     if (adj) then
       if (.not. add) then
        select type (flt3d)
        type is (lv_filter1d)
        flt3d%coef1=0.
        type is (lv_filter2d)
        flt3d%coef2=0.
        type is (lv_filter3d)
        flt3d%coef3=0.
        end select
       end if
       call adjoint3d_m(model,flt3d,data)
     else
       if (.not. add) data=0.
       call forward3d(model,flt3d,data)
     end if  
  end subroutine lv_conv3d_m
  
  subroutine lv_conv3d_f(adj,add,mod, data)
     logical, intent(in)     :: adj, add
     real, dimension(:,:,:)		 :: mod, data
     
     if (adj) then
       if (.not. add) mod=0.
       call adjoint3d_f(mod,flt,data)
     else
       if (.not. add) data=0.
       call forward3d(mod,flt,data)
     end if  
  end subroutine lv_conv3d_f  
  
  subroutine lv_conv3d_close_m()
    nullify(model)
  end subroutine lv_conv3d_close_m
  
  subroutine lv_conv3d_close_f()
    nullify(flt)
  end subroutine lv_conv3d_close_f
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine forward3d(m,f,d) !output d=f*m
     real, dimension(:,:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:,:)		:: d
     
     integer        :: i,j,id1,id2,id3,k1,k2,k3,nlag

     real, allocatable :: d1(:), d21(:), d22(:), d31(:), d32(:), d33(:), d34(:)
     real, allocatable :: b1(:), b21(:), b22(:), b31(:), b32(:), b33(:), b34(:)
     
     select type (f)
     
     type is (lv_filter1d)
!    
     type is (lv_filter2d)
!    
     type is (lv_filter3d)

     nlag=size(f%lag,2)
     allocate(d1(nlag), d21(nlag), d22(nlag), d31(nlag), d32(nlag), d33(nlag), d34(nlag))
     allocate(b1(nlag), b21(nlag), b22(nlag), b31(nlag), b32(nlag), b33(nlag), b34(nlag))
     
    !$OMP PARALLEL DO PRIVATE(i,k1,k2,k3,b31,b32,b33,b34,b21,b22,b1,d31,d32,d33,d34,d21,d22,d1,id3,id2,id1,j)
     do i=1,size(f%b,2)  									! i-th patch
       k1 = mod(mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)),(size(f%coef3,2)-1)) +1
       k2 = (mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)))/(size(f%coef3,2)-1) +1
       k3 = (i-1)/((size(f%coef3,2)-1)*(size(f%coef3,3)-1))+1
       b31= f%coef3(:,k1,k2,k3)
       b32= f%coef3(:,k1+1,k2,k3)
       b33= f%coef3(:,k1+1,k2+1,k3)
       b34= f%coef3(:,k1,k2+1,k3)
       d31= (f%coef3(:,k1,k2,k3+1)-b31)/(f%e(3,i)-f%b(3,i)+1)
       d32= (f%coef3(:,k1+1,k2,k3+1)-b32)/(f%e(3,i)-f%b(3,i)+1)
       d33= (f%coef3(:,k1+1,k2+1,k3+1)-b33)/(f%e(3,i)-f%b(3,i)+1)
       d34= (f%coef3(:,k1,k2+1,k3+1)-b34)/(f%e(3,i)-f%b(3,i)+1)
       do id3=f%b(3,i),f%e(3,i) 							! 3rd axis
         b21=b31
         b22=b32
         d21=(b34-b31)/(f%e(2,i)-f%b(2,i)+1)
         d22=(b33-b32)/(f%e(2,i)-f%b(2,i)+1)
       	 do id2=f%b(2,i),f%e(2,i)  							! 2rd axis
           b1=b21
           d1=(b22-b21)/(f%e(1,i)-f%b(1,i)+1)
           do id1=f%b(1,i),f%e(1,i) 						! 1st axis
         ! TODO: mask check (mask is 2D)
             do j = 1,nlag
               d(id1+f%lag(1,j),id2+f%lag(2,j),id3+f%lag(3,j))= &
               d(id1+f%lag(1,j),id2+f%lag(2,j),id3+f%lag(3,j))+b1(j)*m(id1,id2,id3)
             end do       
             b1=b1+d1       
           end do   
           b21=b21+d21
           b22=b22+d22  
      	  end do ! end of 2nd axis loop
    	  b31=b31+d31
    	  b32=b32+d32
    	  b33=b33+d33
    	  b34=b34+d34
     	end do ! end of 3 axis loop
     end do ! end of patch loop
	!$OMP END PARALLEL DO 
     
     deallocate(d1, d21, d22, d31, d32, d33, d34, b1, b21, b22, b31, b32, b33, b34)
     
     end select
     
  end subroutine forward3d

  subroutine adjoint3d_f(m,f,d)
     real, dimension(:,:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:,:)		:: d     
     
     integer        :: i,j,id1,id2,id3,k1,k2,k3,nlag

     real, allocatable :: d1(:), d21(:), d22(:), d31(:), d32(:), d33(:), d34(:)
     real, allocatable :: b1(:), b21(:), b22(:), b31(:), b32(:), b33(:), b34(:)
     
     select type (f)
     
     type is (lv_filter1d)
!    
     type is (lv_filter2d)
!    
     type is (lv_filter3d)

     nlag=size(f%lag,2)
     allocate(d1(nlag), d21(nlag), d22(nlag), d31(nlag), d32(nlag), d33(nlag), d34(nlag))
     allocate(b1(nlag), b21(nlag), b22(nlag), b31(nlag), b32(nlag), b33(nlag), b34(nlag))
     
     !$OMP PARALLEL DO PRIVATE(i,k1,k2,k3,b31,b32,b33,b34,b21,b22,b1,d31,d32,d33,d34,d21,d22,d1,id3,id2,id1,j)
     do i=1,size(f%b,2)  									! i-th patch
       k1 = mod(mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)),(size(f%coef3,2)-1)) +1
       k2 = (mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)))/(size(f%coef3,2)-1) +1
       k3 = (i-1)/((size(f%coef3,2)-1)*(size(f%coef3,3)-1))+1
       b31= f%coef3(:,k1,k2,k3)
       b32= f%coef3(:,k1+1,k2,k3)
       b33= f%coef3(:,k1+1,k2+1,k3)
       b34= f%coef3(:,k1,k2+1,k3)
       d31= (f%coef3(:,k1,k2,k3+1)-b31)/(f%e(3,i)-f%b(3,i)+1)
       d32= (f%coef3(:,k1+1,k2,k3+1)-b32)/(f%e(3,i)-f%b(3,i)+1)
       d33= (f%coef3(:,k1+1,k2+1,k3+1)-b33)/(f%e(3,i)-f%b(3,i)+1)
       d34= (f%coef3(:,k1,k2+1,k3+1)-b34)/(f%e(3,i)-f%b(3,i)+1)
       do id3=f%b(3,i),f%e(3,i) 							! 3rd axis
         b21=b31
         b22=b32
         d21=(b34-b31)/(f%e(2,i)-f%b(2,i)+1)
         d22=(b33-b32)/(f%e(2,i)-f%b(2,i)+1)
       	 do id2=f%b(2,i),f%e(2,i)  							! 2rd axis
           b1=b21
           d1=(b22-b21)/(f%e(1,i)-f%b(1,i)+1)
           do id1=f%b(1,i),f%e(1,i) 						! 1st axis
         ! TODO: mask check (mask is 2D)
             do j = 1,nlag
               m(id1,id2,id3)= m(id1,id2,id3) &
               + b1(j)*d(id1+f%lag(1,j),id2+f%lag(2,j),id3+f%lag(3,j))
             end do       
             b1=b1+d1       
           end do   
           b21=b21+d21
           b22=b22+d22  
      	  end do ! end of 2nd axis loop
    	  b31=b31+d31
    	  b32=b32+d32
    	  b33=b33+d33
    	  b34=b34+d34
     	end do ! end of 3 axis loop
     end do ! end of patch loop
	 !$OMP END PARALLEL DO 
     
     deallocate(d1, d21, d22, d31, d32, d33, d34, b1, b21, b22, b31, b32, b33, b34)
     
     end select

  end subroutine adjoint3d_f
  
  subroutine adjoint3d_m(m,f,d)
     real, dimension(:,:,:)		:: m
     class(lv_filter)			:: f
     real, dimension(:,:,:)		:: d
     
     integer        :: i,j,id1,id2,id3,k1,k2,k3,nlag,np1,np2,np3,i3,i2,i1,k

     real, allocatable :: d1(:), d21(:), d22(:), d31(:), d32(:), d33(:), d34(:)
     real, allocatable :: b1(:), b21(:), b22(:), b31(:), b32(:), b33(:), b34(:)
     
     double precision         :: invn1, invn2, invn3
     real :: x
     
     select type (f)
     
     type is (lv_filter1d)
     !
     type is (lv_filter2d)
     
     type is (lv_filter3d)
     nlag=size(f%lag,2)

     allocate(d1(nlag), d21(nlag), d22(nlag), d31(nlag), d32(nlag), d33(nlag), d34(nlag))
     allocate(b1(nlag), b21(nlag), b22(nlag), b31(nlag), b32(nlag), b33(nlag), b34(nlag))
     
     np1=size(f%coef3,2)-1
     np2=size(f%coef3,3)-1
     np3=size(f%coef3,4)-1
     
     !Call OMP_SET_NUM_Threads(10) 
     
 !    do i3=1,np3/2
!     	do i2=1,np2/2
!     	  do i1=1,np1/2
!     	     do k=1,8
!     	    ! goal: to find i
!     	     i=(i3*2-2)*np2*np1+ (i2*2-2)*np1+(i1-1)*2+np1*np2*(k-1)/4 + mod((k-1)/2,2)*np1+(mod(k-1,2)+1)

	!$OMP PARALLEL DO PRIVATE(i,k1,k2,k3,b31,b32,b33,b34,b21,b22,b1,d31,d32,d33,d34,d21,d22,d1,id3,id2,id1,j,invn1,invn2,invn3)
     do i=1,size(f%b,2)  ! i-th patch
       invn1=1./(f%e(1,i)-f%b(1,i)+1)
       invn2=1./(f%e(2,i)-f%b(2,i)+1)
       invn3=1./(f%e(3,i)-f%b(3,i)+1)
       k1 = mod(mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)),(size(f%coef3,2)-1)) +1
       k2 = (mod(i-1,(size(f%coef3,2)-1)*(size(f%coef3,3)-1)))/(size(f%coef3,2)-1) +1
       k3 = (i-1)/((size(f%coef3,2)-1)*(size(f%coef3,3)-1))+1
      
       d31=0
       d32=0
       d33=0
       d34=0
       b31=0
       b32=0
       b33=0
       b34=0
       do id3=f%e(3,i),f%b(3,i),-1 
 	     d21=0
	     d22=0
         b21=0
         b22=0
         do id2=f%e(2,i),f%b(2,i),-1
           b1=0
           d1=0
           do id1=f%e(1,i),f%b(1,i),-1
         ! TODO: mask check (mask is 2D)
             do j = 1,nlag
                x= m(id1,id2,id3)*d(id1+f%lag(1,j),id2+f%lag(2,j),id3+f%lag(3,j))
                d1(j)=d1(j)+b1(j)
            	b1(j)=x+b1(j)
             end do 						! end of lag loop
            ! d1=d1+b1
           end do  							! end of 1st axis loop
           d21=d21+b21
           d22=d22+b22
           b21 = b21+b1-d1*invn1
           b22 = b22+d1*invn1             
		 end do 							! end of 2 axis  
         d31=d31+b31
         d32=d32+b32
         d33=d33+b33
         d34=d34+b34
         b31=b31+b21-invn2*d21
         b32=b32+b22-invn2*d22
         b33=b33+invn2*d22
         b34=b34+invn2*d21  
       end do 								! end of 3 axis
       
       f%coef3(:,k1,k2,k3)      = f%coef3(:,k1,k2,k3)+b31-invn3*d31
       f%coef3(:,k1+1,k2,k3)    = f%coef3(:,k1+1,k2,k3)+b32-invn3*d32
       f%coef3(:,k1+1,k2+1,k3)  = f%coef3(:,k1+1,k2+1,k3)+b33-invn3*d33
       f%coef3(:,k1,k2+1,k3)    = f%coef3(:,k1,k2+1,k3)+b34-invn3*d34
       f%coef3(:,k1,k2,k3+1)    = f%coef3(:,k1,k2,k3+1)+invn3*d31
       f%coef3(:,k1+1,k2,k3+1)  = f%coef3(:,k1+1,k2,k3+1)+invn3*d32
       f%coef3(:,k1+1,k2+1,k3+1)= f%coef3(:,k1+1,k2+1,k3+1)+invn3*d33
       f%coef3(:,k1,k2+1,k3+1)  = f%coef3(:,k1,k2+1,k3+1)+invn3*d34
     end do 
     !$OMP END PARALLEL DO 	
 !    end do
!     end do
!     end do
     
     deallocate(d1, d21, d22, d31, d32, d33, d34, b1, b21, b22, b31, b32, b33, b34)
     
     end select
   
  end subroutine adjoint3d_m



end module
