module solver_reg_mod 
! 0 = F m - d and 0 = A m
  implicit none
  contains
  subroutine solver_reg(m,d, Fop, Aop, stepper, nAop, niter,eps,m0,rm0,resd,resm)
    optional :: m0,rm0,resd,resm
    interface  
!-------------------------- begin definitions -----------
      integer function Fop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Aop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function stepper(forget,m,g,rr,gg)
        real, dimension(:)  ::        m,g,rr,gg    
        logical             :: forget
      end function
    end interface 
    real, dimension(:),    intent(in)          :: d, m0,rm0
    integer,               intent(in)          :: niter, nAop
    real,                  intent(in)          :: eps
    real, dimension(:),    intent(out)         :: m, resd,resm
    real, dimension(size( m))                  :: g
    real, dimension(size( d) + nAop), target   :: rr, gg
    real, dimension(:), pointer                :: rd, gd
    real, dimension(:), pointer                :: rm, gm
    integer                                    :: iter, stat
    logical                                    :: forget
    rd => rr(1:size(d))
    rm => rr(1+size(d):)
    gd => gg(1:size(d))
    gm => gg(1+size(d):)

    rd = -d                                !Rd = -d
    rm = 0.
    if (present(rm0)) then
      rm=rm0                               !Rm = Rm0
    end if
    if (present( m0)) then
      m=m0                                 !m   = m0
      stat = Fop(.false.,.true.,m,rd)      !Rd += F m0
      stat = Aop(.false.,.true.,eps*m0,rm) !Rm += e A m0
    else
      m=0
    end if 
    write(0,*) "       iter      rd*rd          rm*rm"
!--------------------------- begin iterations -----------
    do iter = 1,niter  
      stat = Fop(.true.,.false.,g,rd)       !g  = (F)'Rd
      stat = Aop(.true.,.true.,g,eps*rm)    !g += e A'Rm
      stat = Fop(.false.,.false.,g,gd)      !Gd = (F) g
      stat = Aop(.false.,.false.,eps*g,gm)  !Gm = e A g
      stat = stepper(forget, m,g, rr,gg)    !m+=dm; R+=dR
      if (stat .eq.1) then
        exit ! got stuck descending
      end if
      write(0,*) iter, sum(rd*rd), sum(rm*rm)
      if(sum(rr*rr) < 1.e-30) then
        write(0,*) "residual norm is too small. Will stop iterating"
        exit
      end if
      if(sum(g*g) < 1.e-30) then
        write(0,*) "gradient norm is too small. Will stop iterating"
        exit
      end if
      if(sum(gg*gg) < 1.e-30) then
        write(0,*) "residual update norm is too small. Will stop iterating"
        exit
      end if
    end do 
    if (present( resd)) then
      resd = rd
    end if
    if (present( resm)) then
      resm = rm(:size(resm))
    end if
  end subroutine 
end module 
